'''
HDLC-like communications
'''

__version__ = '0.7.0'

import logging
import struct
import time
from enum import Enum
from typing import Callable, Union

logger = logging.getLogger(__name__)

class HDLCStats:
    '''
    HDLCStats - statistics on HDLC-like communication
    '''
    crcErrors = 0

class HDLCState(Enum):
    STATE_READ = 1
    STATE_ESCAPE = 2

# Frame
class Frame(object):
    '''
    Frame - represents an encoded HDLC-like frame
    '''
    CRC16_LUT = [
        0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
        0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
        0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
        0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
        0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
        0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
        0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
        0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
        0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
        0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
        0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
        0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
        0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
        0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
        0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
        0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
        0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
        0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
        0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
        0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
        0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
        0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
        0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
        0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
        0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
        0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
        0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
        0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
        0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
        0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
        0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
        0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
        ]

    DELIMITER_CODE_NON_ASCII = 0xE7
    ESCAPE_CODE_NON_ASCII = 0xD7
    DELIMITER_CODE_ASCII = 0x7E
    ESCAPE_CODE_ASCII = 0x7D

    def __init__(self, delimiterCode: int, escapeCode: int) -> None:
        '''
        Initialise Frame

        Args:
            asciiEscapes: bool - use the standard HDLC escape codes (which are ASCII values)

        Returns:
            None
        '''
        self.finished = False
        self.error = False
        self.state = HDLCState.STATE_READ
        self.data = bytearray()
        self.crc = bytearray()
        self.reader = None
        self.delimiterCode = delimiterCode
        self.escapeCode = escapeCode

    def __len__(self) -> int:
        return len(self.data)

    def addByte(self, b: int) -> None:
        '''
        Add byte to frame

        Args:
            b: byte to add

        Returns:
            None
        '''
        if b == self.escapeCode:
            self.state = HDLCState.STATE_ESCAPE
        elif self.state == HDLCState.STATE_ESCAPE:
            self.state = HDLCState.STATE_READ
            b = b ^ 0x20
            self.data.append(b)
        else:
            self.data.append(b)

    def finish(self) -> None:
        '''
        Complete frame creation - updates crc and data variables

        Args:
            none

        Returns:
            None
        '''
        self.crc = self.data[-2:]
        self.data = self.data[:-2]
        self.finished = True

    def checkCRC(self) -> bool:
        '''
        Check the CRC received matches expected

        Args:
            none

        Returns:
            bool - False indicates CRC check failure
        '''
        res = bool(self.crc == LikeHDLC.calcCRC(self.data))
        if not res:
            logger.warning(f"invalid crc {self.crc} != {LikeHDLC.calcCRC(self.data)}")
            self.error = True
        return res
        
    def toString(self) -> str:
        return self.data.decode('utf-8').rstrip('\0')

# LikeHDLC
class LikeHDLC:
    '''
    HDLC-like with encode and decode capabilities
    To decode, construct LikeHDLC and supply onFrame callback -
        then send data to decodeData() and onFrame
        callback will be called for each frame decoded
    To encode use class method encode()
    The communications protocol is based on HDLC but with some simplifications:
    - byte-wise only, escape codes can be modified from standard
    - encoding and decoding only
    '''
    def __init__(self, 
            onFrame: Callable[[Union[bytes, str]], None],
            onError: Callable[[], None],
            asciiEscapes: bool = False,
            payloadsAreStrings: bool = False) -> None:
        '''
        Initialise LikeHDLC
        Args:
            onFrame: callback function (takes 1 parameter which is a received frame)
            onError: callback function (0 parameters) when an error occurs
            asciiEscapes: bool - use the standard HDLC escape codes (which are ASCII values)
            payloadsAreStrings: bool - received frames are returned as strings (as opposed to bytes)
        Returns:
            None
        '''
        self.payloadsAreStrings = payloadsAreStrings
        self.onFrame = onFrame
        self.onError = onError
        self.asciiEscapes = asciiEscapes
        self.delimiterCode = Frame.DELIMITER_CODE_ASCII if asciiEscapes \
                        else Frame.DELIMITER_CODE_NON_ASCII
        self.escapeCode = Frame.ESCAPE_CODE_ASCII if asciiEscapes \
                        else Frame.ESCAPE_CODE_NON_ASCII
        self.clear()
        self.clearStats()

    def clear(self) -> None:
        
        '''
        Get statistics on HDLC conversion
        Args:
            none
        Returns:
            None
        '''
        self.currentFrame = None

    def clearStats(self) -> None:
        '''
        Clear statistics on HDLC operation
        Args:
            none
        Returns:
            None
        '''
        self.stats = HDLCStats()

    def decodeData(self, b: int) -> None:
        '''
        Add a byte to be decoded
        Args:
            b byte to add
        Returns:
            None
        '''
        if b < 0 or b > 255:
            return
        if b == self.delimiterCode:
            # Start or End
            if (self.currentFrame is None) or len(self.currentFrame) < 1:
                # Start
                self.currentFrame = Frame(self.delimiterCode, self.escapeCode)
                # logger.debug(f"START time {time.time()}")
            else:
                # End
                self.currentFrame.finish()
                self.currentFrame.checkCRC()
                # logger.debug(f"END time {time.time()}")
        else:
            if self.currentFrame is not None:
                if not self.currentFrame.finished:
                    # Add to current frame
                    self.currentFrame.addByte(b)

        # Validate and callback if finished
        if (self.currentFrame is not None) and self.currentFrame.finished:
            if not self.currentFrame.error:
                # Success
                if self.payloadsAreStrings:
                    self.onFrame(self.currentFrame.toString())
                else:
                    self.onFrame(self.currentFrame.data)
            elif self.currentFrame.finished:
                # Error
                self.stats.crcErrors += 1
                self.onError()
            self.currentFrame = None

    def getStats(self) -> HDLCStats:
        
        '''
        Get statistics on HDLC conversion

        Args:
            none

        Returns:
            HDLCStats object containing statistics
        '''
        return self.stats

    @classmethod
    def encode(cls, data: bytes, asciiEscapes: bool = False) -> bytes:
        '''
        Encode data into an HDLC-like frame

        Args:
            data: bytes representing the data

        Returns:
            bytes representing the frame
        '''
        delimiterCode = Frame.DELIMITER_CODE_ASCII if asciiEscapes \
                    else Frame.DELIMITER_CODE_NON_ASCII
        escapeCode = Frame.ESCAPE_CODE_ASCII if asciiEscapes \
                    else Frame.ESCAPE_CODE_NON_ASCII
        frame = bytearray()
        frame.append(delimiterCode)
        crc = cls.calcCRC(data)
        data = data + crc
        for byte in data:
            if byte == delimiterCode or byte == escapeCode:
                frame.append(escapeCode)
                frame.append(byte ^ 0x20)
            else:
                frame.append(byte)
        frame.append(delimiterCode)
        return bytes(frame)

    @classmethod
    def calcCRC(cls, data: bytes) -> bytes:
        
        '''
        Calculate CRC-CCITT 16Bit check code
        https://en.wikipedia.org/wiki/High-Level_Data_Link_Control

        Args:
            data: bytes representing the data

        Returns:
            bytes representing the checksum (2 bytes big-endian)
        '''
        
        crc = 0xffff
        for c in data:
            crc = ((crc<<8)&0xff00) ^ Frame.CRC16_LUT[((crc>>8)&0xff)^c]
        return bytearray(struct.pack(">H", crc))

    @classmethod
    def toBytes(cls, data):
        '''
        Convert iterable data (list, etc) into bytes

        Args:
            data: the data to turn into bytes (1 byte is added per value)

        Returns:
            None
        '''
        return bytearray(data)



from typing import Callable, Dict, List, Tuple, Union
import sys
import time
import socket
import struct
from .ClientGeneric import ClientGeneric
from .Exceptions import (MartyConnectException,
                         MartyCommandException,
                         ArgumentOutOfRangeException,
                         UnavailableCommandException)

class ClientMV1(ClientGeneric):
    '''
    Lower level interface class between the `Marty` abstracted
    control class and the RIC socket interface
    '''
    SOCKET_PORT = 24

    SIDE_CODES = {
        'left'    : '\x00',
        'right'   : '\x01',
        'forward' : '\x02',
        'back'    : '\x03',
        'auto'    : '\xff',
    }

    EYE_POSES = {
        'angry': -15,
        'excited': 30,
        'normal': 0,
        'wide': 90
    }

    def __init__(self, method: str, locator: str,
                port:str = None, timeout:float = 5.0,
                debug:bool = False,
                default_lifelike: bool = True,
                *args, **kwargs):
        '''
        Initialise connection to remote Marty over a IPv4 socket by name 'loc' over port 24
        Args:
            loc, type str, must either resolve to an IP or be an IP address
        Raises:
            MartyConnectException if the socket failed to make the connection to the host
        '''
        if port is None:
            self.port = self.SOCKET_PORT
        else:
            self.port = port

        self.debug = debug

        self.loc = locator
        self.timeout = timeout
        self.default_lifelike = default_lifelike

        self.sock = self._socket_factory()

        self.COMMANDS_LUT = dict()
        self._init_commands()


    def start(self):
        # To be able to do anything:
        self.enable_safeties(True)
        self.enable_motors(True)

        if self.default_lifelike:
            self.lifelike_behaviour(True)

    def close(self):
        if self.sock:
            self.sock.close()
        self.sock = None

    def hello(self):
        return self._execute('hello')

    def get_ready(self):
        return self._execute('hello')

    def discover(self):
        return self._discover()

    def stop(self, stop_type: str, stopCode: int) -> bool:
        M1_STOP_TYPE = ['\x00','\x01','\x02','\x03','\x04','\x05']
        return self._execute('stop', M1_STOP_TYPE[stopCode])

    def resume(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def hold_position(self, hold_time: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def move_joint(self, joint_id: int, position: int, move_time: int) -> bool:
        '''
        Move a specific joint to a position
        Args:
            move_time: how long this movement should last, in milliseconds
        '''
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('move_joint',
                                   self._pack_uint8(joint_id),
                                   self._pack_int8(position),
                                   dur_lsb, dur_msb)

    def get_joint_position(self, joint_id: Union[int, str]) -> float:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_joint_current(self, joint_name_or_num: Union[int, str]) -> float:
        if type(joint_name_or_num) is str:
            raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)
        return self.get_motor_current(int(joint_name_or_num))

    def get_joint_status(self, joint_name_or_num: Union[int, str]) -> int:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def lean(self, direction: str, amount: int, move_time: int) -> bool:
        try:
            directionNum = self.SIDE_CODES[direction]
        except KeyError:
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(self.SIDE_CODES.keys()), direction))
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('lean', directionNum,
                                   self._pack_int8(amount),
                                   dur_lsb, dur_msb)

    def walk(self, num_steps: int = 2, start_foot:str = 'auto', turn: int = 0,
                step_length:int = 40, move_time: int = 1500) -> bool:
        try:
            side_c = self.SIDE_CODES[start_foot]
        except KeyError:
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(self.SIDE_CODES.keys()), start_foot))
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('walk',
                                   self._pack_uint8(num_steps),
                                   self._pack_int8(turn),
                                   dur_lsb, dur_msb,
                                   self._pack_int8(step_length),
                                   side_c)

    def eyes(self, joint_id: int, pose_or_angle: Union[str, int], move_time: int = 100) -> bool:
        if type(pose_or_angle) is str:
            try:
                angle = self.EYE_POSES[pose_or_angle]
            except KeyError:
                raise MartyCommandException("pose must be one of {}, not '{}'"
                                            "".format(set(self.EYE_POSES.keys()), pose_or_angle))
        else:
            angle = pose_or_angle
        return self.move_joint(joint_id, angle, move_time)

    def kick(self, side: str = 'right', twist: int = 0, move_time: int = 2000) -> bool:
        side_c = self.SIDE_CODES[side]
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('kick', side_c,
                                   self._pack_int8(twist),
                                   dur_lsb, dur_msb)

    def arms(self, left_angle: int, right_angle: int, move_time: int) -> bool:
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('arms',
                                   self._pack_int8(right_angle),
                                   self._pack_int8(left_angle),
                                   dur_lsb, dur_msb)

    def celebrate(self, move_time: int = 4000) -> bool:
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('celebrate', dur_lsb, dur_msb)

    def circle_dance(self, side: str = 'right', move_time: int = 1500) -> bool:
        side_c = self.SIDE_CODES[side]
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('circle_dance',
                                   side_c,
                                   dur_lsb, dur_msb)

    def dance(self, side: str = 'right', move_time: int = 1500) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def wiggle(self, move_time: int = 1500) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def sidestep(self, side: str, steps: int = 1, step_length: int = 100,
            move_time: int = 2000) -> bool:
        side_c = self.SIDE_CODES[side]
        dur_lsb, dur_msb = self._pack_uint16(move_time)
        return self._execute('sidestep', side_c,
                                   self._pack_int8(steps),
                                   dur_lsb, dur_msb,
                                   self._pack_int8(step_length))

    def play_sound(self, freq_start: Union[float, str],
            freq_end: float,
            duration: int) -> bool:
        if type(freq_start) is str:
            raise MartyCommandException("first parameter must be a number for Marty V1")
        f_start_lsb, f_start_msb = self._pack_uint16(int(freq_start))
        f_end_lsb, f_end_msb = self._pack_uint16(int(freq_end))
        dur_lsb, dur_msb = self._pack_uint16(5000 if duration is None else duration)
        return self._execute('play_sound',
                                   f_start_lsb, f_start_msb,
                                   f_end_lsb, f_end_msb,
                                   dur_lsb, dur_msb)

    GPIO_PIN_MODES = {
        'digital in'  : '\x00',
        'analog in'   : '\x01',
        'digital out' : '\x02',
        #'servo'       : '\x03',
        #'pwm'         : '\x04',
    }

    def pinmode_gpio(self, gpio: int, mode: str) -> bool:
        return self._execute('gpio_mode', self._pack_uint8(gpio), self.GPIO_PIN_MODES[mode])

    def write_gpio(self, gpio: int, value: int) -> bool:
        byte_array = (self._pack_uint8(gpio),) + self._pack_float(value)
        return self._execute('gpio_write', *byte_array)

    def digitalread_gpio(self, gpio: int) -> bool:
        return bool(self._execute('gpio', self._pack_uint8(gpio)))

    def i2c_write(self, *byte_array: int) -> bool:
        return self._execute('i2c_write', *byte_array)

    def i2c_write_to_ric(self, address: int, byte_array: bytes) -> bool:
        data = ['\x1B']         #i2c_write opcode
        data.append(address)    #i2c address
        data.append(byte_array) #message
        data += self.ros_serial_formatter(111, *byte_array) #ros serial format
        return self._execute('i2c_write', *data)

    def get_battery_voltage(self) -> float:
        return self._execute('battery')

    def get_battery_remaining(self) -> float:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_distance_sensor(self) -> float:
        return self._execute('distance')

    def get_accelerometer(self, axis: str, axisCode: int) -> float:
        M1_ACCEL_AXES = ['\x00','\x01','\x02']
        if axis is None:
            raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)
        return self._execute('accel', M1_ACCEL_AXES[axisCode])

    def get_motor_current(self, motor_id: int) -> float:
        return self._execute('motorcurrent', str(motor_id))

    def enable_motors(self, enable: bool = True, clear_queue: bool = True) -> bool:
        if clear_queue:
            self.stop('clear queue', 0)
        if enable:
            return self._execute('enable_motors') and True
        else:
            return self._execute('disable_motors') and False

    def enable_safeties(self, enable: bool = True) -> bool:
        return self._execute('enable_safeties')

    def fall_protection(self, enable: bool = True) -> bool:
        return self._execute('fall_protection', enable)

    def motor_protection(self, enable: bool = True) -> bool:
        return self._execute('motor_protection', enable)

    def battery_protection(self, enable: bool = True) -> bool:
        return self._execute('battery_protection', enable)

    def buzz_prevention(self, enable: bool = True) -> bool:
        return self._execute('buzz_prevention', enable)

    def lifelike_behaviour(self, enable: bool = True) -> bool:
        return self._execute('lifelike_behaviour', enable)

    def set_parameter(self, *byte_array: int) -> bool:
        return self._execute('set_param', '\x1F', *byte_array)

    def save_calibration(self) -> bool:
        return self._execute('save_calibration')

    def clear_calibration(self) -> bool:
        return self._execute('clear_calibration')

    def is_calibrated(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def ros_command(self,  *byte_array: int) -> bool:
        return self._execute('ros_command', *byte_array)

    def keyframe (self, time: float, num_of_msgs: int, msgs: List) -> List[bytes]:
        processed_keyframe = []

        #Number of key frames
        len_byte0, len_byte1, len_byte2, len_byte3 = self._pack_int32(1)
        processed_keyframe.append(len_byte0)
        processed_keyframe.append(len_byte1)
        processed_keyframe.append(len_byte2)
        processed_keyframe.append(len_byte3)

        #Time (in seconds) to excute keyframe. This is float encoded.
        time_byte0, time_byte1, time_byte2, time_byte3 = self._pack_float(time)
        processed_keyframe.append(time_byte0)
        processed_keyframe.append(time_byte1)
        processed_keyframe.append(time_byte2)
        processed_keyframe.append(time_byte3)

        if(len(msgs) != num_of_msgs):
            raise MartyCommandException('Number of messages do not match entered messages')
        #Array length
        arr_len_byte0, arr_len_byte1, arr_len_byte2, arr_len_byte3 = self._pack_int32(num_of_msgs)
        processed_keyframe.append(arr_len_byte0)
        processed_keyframe.append(arr_len_byte1)
        processed_keyframe.append(arr_len_byte2)
        processed_keyframe.append(arr_len_byte3)

        #Messages
        for items in msgs:
            for values in items:
                processed_keyframe.append(self._pack_int8(values))

        return(processed_keyframe)

    def get_chatter(self) -> bytes:
        return self._execute('chatter')

    def get_firmware_version(self) -> bool:
        return self._execute('firmware_version')

    def _mute_serial(self) -> bool:
        return self._execute('mute_serial')

    def ros_serial_formatter(self, topicID: int, send: bool = False, *message: int) -> List[int]:
        msg = message

        msg_length = len(msg)
        #Message length in little endian format
        msg_length_LB = msg_length & 0xFF                           #3rd byte
        msg_length_HB = (msg_length >> 8) & 0xFF                    #4th byte

        checksum1 = 255 - ((msg_length_LB + msg_length_HB) % 256)   #5th byte

        #Topic ID in little endian format
        topic_ID_LB = topicID & 0xFF                                #6th byte
        topic_ID_HB = (topicID >> 8) & 0xFF                         #7th byte

        data_values_sum = 0
        for i in msg:
            data_values_sum += ord(i)

        checksum2 = 255 - ((topic_ID_LB + topic_ID_HB + data_values_sum) % 256) #final byte

        #encode into bytes
        command_to_be_sent = []
        command_to_be_sent += ('\xff',) #Sync Flag. Check ROS Wiki
        command_to_be_sent += ('\xfe',) #Protocol version. Check ROS Wiki
        command_to_be_sent += (chr(msg_length_LB),)
        command_to_be_sent += (chr(msg_length_HB),)
        command_to_be_sent += (chr(checksum1),)
        command_to_be_sent += (chr(topic_ID_LB),)
        command_to_be_sent += (chr(topic_ID_HB),)
        command_to_be_sent += msg
        command_to_be_sent += (chr(checksum2),)

        if(send == True):
            self.ros_command(*command_to_be_sent)

        return(command_to_be_sent)

    def _init_commands(self) -> None:

        # Extend basis LUT with specific handlers
        self._register_commands({
            'discover'           : self._discover,
            'battery'            : self._simple_sensor,
            'distance'           : self._simple_sensor,
            'accel'              : self._select_sensor,
            'motorcurrent'       : self._current_sensor,
            'gpio'               : self._select_sensor,
            'hello'              : self._fixed_command,
            'lean'               : self._fixed_command,
            'walk'               : self._fixed_command,
            'kick'               : self._fixed_command,
            'celebrate'          : self._fixed_command,
            'arms'               : self._fixed_command,
            'sidestep'           : self._fixed_command,
            'circle_dance'       : self._fixed_command,
            'play_sound'         : self._fixed_command,
            'stop'               : self._fixed_command,
            'move_joint'         : self._fixed_command,
            'enable_motors'      : self._fixed_command,
            'disable_motors'     : self._fixed_command,
            'enable_safeties'    : self._fixed_command,
            'fall_protection'    : self._toggle_command,
            'motor_protection'   : self._toggle_command,
            'battery_protection' : self._toggle_command,
            'buzz_prevention'    : self._toggle_command,
            'lifelike_behaviour' : self._toggle_command,
            'clear_calibration'  : self._fixed_command,
            'save_calibration'   : self._fixed_command,
            'ros_command'        : self._command,
            'chatter'            : self._chatter,
            'set_param'          : self._command,
            'firmware_version'   : self._fixed_command,
            'mute_serial'        : self._fixed_command,
            'i2c_write'          : self._command,
            'gpio_write'         : self._fixed_command,
            'gpio_mode'          : self._fixed_command,
        })

        # for opcode, _ in self.COMMANDS_LUT.items():
        #     if opcode not in self.CMD_OPCODES.keys():
        #         print('Missing {}'.format(opcode))

    def _socket_factory(self) -> socket.socket:
        '''
        Generate a new socket connection
        '''
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            loc_ip = socket.gethostbyname(self.loc)
            sock.connect((loc_ip, self.port))
        except Exception as e:
            raise MartyConnectException(e)

        sock.settimeout(self.timeout)

        if sys.platform == 'linux':
            # Configure socket in keepalive mode
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # Time to wait before sending a keepalive
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 1)
            # Interval between keepalives
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 1)
            # Number of keepalive fails before declaring the connection broken
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 5)

        elif sys.platform == 'windows':
            # Keep alive, 5 sec timeout, 1 sec interval
            sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, 5000, 1000))

        return sock

    def __repr__(self):
        return '{}.{}{}'.format(self.__module__,
                             self.__class__.__name__,
                             self.sock.getsockname())

    @staticmethod
    def _discover(timeout: float = 10, *args: int, **kwargs: int) -> List[str]:
        '''
        Search for Marties on the network using a UDB multicast to port 4000
        '''
        socket_addr = "224.0.0.1"
        socket_port = 4000
        magic_command = b"AA"

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 32)
        sock.settimeout(timeout)
        sock.sendto(magic_command, (socket_addr, socket_port))

        found = []
        start = time.time()
        try:
            while (time.time() - start) < timeout:
                data, addr = sock.recvfrom(1000)
                found.append({addr: data})
        except socket.timeout:
            return found

    def is_moving(self) -> bool:
        '''
        Check if Marty is moving

        Args:
            none
        Returns:
            True if Marty is moving
        '''
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def is_paused(self) -> bool:
        '''
        Check if Marty is paused

        Args:
            none
        Returns:
            True if Marty is paused
        '''
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_robot_status(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_joints(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_power_status(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_add_ons_status(self) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_add_on_status(self, add_on_name_or_id: Union[int, str]) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_system_info(self) -> Dict:
        return {"HardwareVersion":"1.0", "SystemName":"MartyV1","SystemVersion":"1.0.0","SerialNo":"000001","MAC":"000000000000"}

    def set_marty_name(self, name: str) -> bool:
        return False

    def get_marty_name(self) -> str:
        return "Marty"

    def is_marty_name_set(self) -> bool:
        return False

    def get_hw_elems_list(self) -> List:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def send_ric_rest_cmd(self, ricRestCmd: str) -> None:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def send_ric_rest_cmd_sync(self, ricRestCmd: str) -> Dict:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    # Encodes Command Type flag, LSB size, MSB size, Data
    CMD_OPCODES = {
        'battery'            : ['\x01', '\x01', '\x00'],         # OK
        'distance'           : ['\x01', '\x08', '\x00'],         #
        'accel'              : ['\x01', '\x02'],                 # OK
        'motorcurrent'       : ['\x01', '\x03'],                 # OK
        'gpio'               : ['\x01', '\x04'],                 #
        'chatter'            : ['\x01', '\x05', '\x00',],        # Variable Length
        'hello'              : ['\x02', '\x01', '\x00', '\x00'], # OK
        'lean'               : ['\x02', '\x05', '\x00', '\x02'], #
        'walk'               : ['\x02', '\x07', '\x00', '\x03'], # OK
        'kick'               : ['\x02', '\x05', '\x00', '\x05'], # OK
        'celebrate'          : ['\x02', '\x03', '\x00', '\x08'], # OK
        'arms'               : ['\x02', '\x05', '\x00', '\x0B'], #
        'sidestep'           : ['\x02', '\x06', '\x00', '\x0E'], #
        'play_sound'         : ['\x02', '\x07', '\x00', '\x10'], #
        'stop'               : ['\x02', '\x02', '\x00', '\x11'], # OK
        'move_joint'         : ['\x02', '\x05', '\x00', '\x12'], #
        'enable_motors'      : ['\x02', '\x01', '\x00', '\x13'], # Has optional args now
        'disable_motors'     : ['\x02', '\x01', '\x00', '\x14'], # Has optional args now
        'circle_dance'       : ['\x02', '\x04', '\x00', '\x1C'], #
        'enable_safeties'    : ['\x02', '\x01', '\x00', '\x1E'], #
        'fall_protection'    : ['\x02', '\x02', '\x00', '\x15'], #
        'motor_protection'   : ['\x02', '\x02', '\x00', '\x16'], #
        'battery_protection' : ['\x02', '\x02', '\x00', '\x17'], #
        'buzz_prevention'    : ['\x02', '\x02', '\x00', '\x18'], #
        'lifelike_behaviour' : ['\x02', '\x02', '\x00', '\x1D'], #
        'clear_calibration'  : ['\x02', '\x01', '\x00', '\xFE'], #
        'save_calibration'   : ['\x02', '\x01', '\x00', '\xFF'], #
        'ros_command'        : ['\x03'],                         # Variable Length
        'set_param'          : ['\x02', '\x1F'], #
        'firmware_version'   : ['\x02', '\x01', '\x00', '\x20'], #
        'mute_serial'        : ['\x02', '\x01', '\x00', '\x21'], # OK
        'i2c_write'          : ['\x02'],                         #
        'gpio_write'         : ['\x02', '\x06', '\x00', '\x1A'], #
        'gpio_mode'          : ['\x02', '\x03', '\x00', '\x19'], #
    }

    def _pack(self, characters: List[str]) -> bytes:
        '''
        Pack characters list into a byte string
        Expects pre-packed chars, use chr or struct.pack to do this
        '''
        if self.debug:
            print(list(map(lambda x: '{}:{}'.format(x, type(x)), characters)))
        try:
            return ("".join(characters)).encode("latin-1")
        except UnicodeEncodeError:
            raise ArgumentOutOfRangeException('Argument(s) overflowed int')

    def _fixed_command(self, *args: int, **kwargs: int) -> bool:
        '''
        Send a command with a fixed (preknown) number of arguments
        that does not expect a response.
        NB!: Assumes data (args[2:]) is broken into 8-bit bytes
        '''
        cmd = args[1]
        opcode = self.CMD_OPCODES[cmd]
        datalen = ord(opcode[1]) + (ord(opcode[2]) << 8) - 1
        data = list(args[2:])
        if len(data) != datalen:
            raise TypeError('{} takes {} arguments but {} were given'
                            ''.format(cmd, datalen, len(data)))
        self.sock.send(self._pack(opcode + data))
        return True

    def _command(self, *args: int, **kwargs: int) -> bool:
        '''
        Pipes args down the socket whilst calculating the payload length
        Args:
            *args of length at least 3
        '''
        cmd = args[1]
        opcode = self.CMD_OPCODES[cmd][0]
        data = list(args[2:])
        datalen_lsb, datalen_msb = struct.pack('<H', len(data))
        payload = [opcode,
                   chr(datalen_lsb[0]),
                   chr(datalen_msb[0])] + data
        self.sock.send(self._pack(payload))
        return True

    def _toggle_command(self, *args: int, **kwargs: int) -> bool:
        '''
        Takes a python Boolean and toggles a switch on the board
        '''
        cmd = args[1]
        toggle = '\x01' if args[2] == True else '\x00'
        opcode = self.CMD_OPCODES[cmd]
        self.sock.send(self._pack(opcode + [toggle]))
        return args[2]

    def _simple_sensor(self, *args: int, **kwargs: int) -> int:
        '''
        Read a simple sensor and give its value
        Args:
            cmd
        '''
        cmd = args[1]
        self.sock.send(self._pack(self.CMD_OPCODES[cmd]))
        data = self.sock.recv(4)
        return struct.unpack('f', data)[0]

    def _select_sensor(self, *args: int, **kwargs: int) -> float:
        '''
        Read a sensor that takes an argument and give its value
        Args:
            cmd, index
        '''
        cmd = args[1]
        index = args[2]
        self.sock.send(self._pack(self.CMD_OPCODES[cmd] + [index]))
        data = self.sock.recv(4)
        return struct.unpack('f', data)[0]

    def _current_sensor(self, *args: int, **kwargs: int) -> float:
        '''
        Read current sensor that takes an argument and give its value
        Args:
            cmd, index
        '''
        cmd = args[1]
        index = args[2]
        pakd = self._pack(self.CMD_OPCODES[cmd] + [chr(int(index))])
        self.sock.send(pakd)
        data = self.sock.recv(4)
        return struct.unpack('f', data)[0]

    def _chatter(self, *args: int, **kwargs: int) -> bytes:
        '''
        Return chatter topic data (variable length)
        '''
        cmd = args[1]
        self.sock.send(self._pack(self.CMD_OPCODES[cmd]))
        data_length = self.sock.recv(1)[0]
        data = self.sock.recv(data_length)
        return data

    def _pack_uint16(self, num: int) -> Tuple[str,str]:
        '''
        Pack an unsigned 16 bit int into two 8 bit bytes, little-endian
        Returns:
            tuple(least-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            h      short/uint16           integer        2
        '''
        try:
            data = struct.pack('<H', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0]), chr(data[1])

    def _pack_int16(self, num: int) -> Tuple[str,str]:
        '''
        Pack a signed 16 bit int into two 8 bit bytes, little-endian
        Returns:
            tuple(least-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            h      short/uint16           integer        2
        '''
        try:
            data = struct.pack('<h', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0]), chr(data[1])

    def _pack_int32(self, num: int) -> Tuple[str,str,str,str]:
        '''
        Pack a signed 32 bit int into four 8 bit bytes, little-endian
        Returns:
            tuple(least-sig-byte, less-sig-byte, more-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            i      int                    integer        4
        '''
        try:
            data = struct.pack('<i', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return (chr(data[0]),
                chr(data[1]),
                chr(data[2]),
                chr(data[3]))

    def _pack_uint8(self, num: int) -> str:
        '''
        Pack an unsigned 8 bit int into one 8 bit byte, little-endian
        Returns:
            bytes

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            B      unsigned char/uint8    integer        1
        '''
        try:
            data = struct.pack('<B', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0])

    def _pack_int8(self, num: int) -> str:
        '''
        Pack a signed 8 bit int into one 8 bit unsigned byte, little-endian
        Returns:
            bytes

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            b      signed char/int8       integer        1
        '''
        try:
            data = struct.pack('<b', num)
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return chr(data[0])

    def _pack_float(self, num: float) -> Tuple[str,str,str,str]:
        '''
        Pack a float into four bit unsigned byte, little-endian
        Returns:
            tuple(least-sig-byte, less-sig-byte, more-sig-byte, most-sig-byte)

        Struct:
            Fmt    C Type                 Python Type    Standard Size
            f      float                  float          4
        '''
        try:
            data = struct.pack('<f', float(num))
        except struct.error as e:
            raise ArgumentOutOfRangeException(e)
        return (chr(data[0]),
                chr(data[1]),
                chr(data[2]),
                chr(data[3]))

    def _execute(self, *args, **kwargs):
        '''
        Execute the command prescribed by the commands lookup table
        '''
        try:
            return self.COMMANDS_LUT[args[0]](self, *args, **kwargs)
        except KeyError:
            raise UnavailableCommandException("The command '{}' is not available "
                                              "from the {} client type"
                                              "".format(args[0], str(self.__class__.__name__)))
        except Exception as e:
            raise MartyCommandException(e)

    def _register_commands(self, handlers):
        '''
        Take {str:func} command names & handlers in a dict
        and register them with the COMMANDS_LUT
        '''
        self.COMMANDS_LUT = ClientGeneric.dict_merge(self.COMMANDS_LUT, handlers)

    def register_logging_callback(self, loggingCallback: Callable[[str],None]) -> None:
        return

    def get_interface_stats(self) -> Dict:
        return {}

    def preException(self, isFatal: bool) -> None:
        pass

    def get_test_output(self) -> dict:
        return ""



import pytest
import time
from martypy import Marty

def test_close() -> None:
    marty = Marty("test", "")
    time.sleep(6)
    marty.close()
    testOutput = marty.get_test_output()
    testExpected = {"buf":
            [
                bytearray(b'\x01\x02\x00v\x00'), 
                bytearray(b'\x02\x02\x00hwstatus\x00'),
                bytearray(b'\x03\x02\x03{"cmdName":"subscription","action":"update","pubRecs":[{"name":"MultiStatus","rateHz":10.0,}{"name":"PowerStatus","rateHz":1.0},{"name":"AddOnStatus","rateHz":10.0}]}\x00\x00'), 
                bytearray(b'\x04\x02\x03{"cmdName":"subscription","action":"update","pubRecs":[{"name":"MultiStatus","rateHz":0},{"name":"PowerStatus","rateHz":0},{"name":"AddOnStatus","rateHz":0}]}\x00\x00')
            ]
        }
    assert(testOutput==testExpected)

def test_circle_dance() -> None:
    marty = Marty("test","",subscribeRateHz=0)
    marty.circle_dance()
    testOutput = marty.get_test_output()
    testExpected = {"buf":
            [
                bytearray(b'\x01\x02\x00v\x00'),
                bytearray(b'\x02\x02\x00hwstatus\x00'),
                bytearray(b'\x03\x02\x00traj/circle?side=0&moveTime=1500\x00')
            ]
        }
    assert(testOutput==testExpected)




from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Optional, Union

class ClientGeneric(ABC):

    SIDE_CODES = {
        'left'    : 1,
        'right'   : 0,
        'forward' : 3,
        'back'    : 2,
        'auto'    : 0,
    }

    EYE_POSES = {
        'angry'   : 'eyesAngry',
        'excited' : 'eyesExcited',
        'normal'  : 'eyesNormal',
        'wide'    : 'eyesWide',
        'wiggle'  : 'wiggleEyes'
    }

    NOT_IMPLEMENTED = "Unfortunately this Marty doesn't do that"

    def __init__(self):
        super().__init__()

    @classmethod
    def dict_merge(cls, *dicts):
        '''
        Merge all provided dicts into one dict
        '''
        merged = {}
        for d in dicts:
            if not isinstance(d, dict):
                raise ValueError('Value should be a dict')
            else:
                merged.update(d)
        return merged

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def hello(self) -> bool:
        return False

    @abstractmethod
    def get_ready(self) -> bool:
        return False

    @abstractmethod
    def discover(self) -> List[str]:
        return []

    @abstractmethod
    def stop(self, stop_type: str, stopCode: int) -> bool:
        return False

    @abstractmethod
    def resume(self) -> bool:
        return False
    
    @abstractmethod
    def hold_position(self, hold_time: int) -> bool:
        return False

    @abstractmethod
    def move_joint(self, joint_id: int, position: int, move_time: int) -> bool:
        return False

    @abstractmethod
    def get_joint_position(self, joint_id: Union[int, str]) -> float:
        return 0

    @abstractmethod
    def get_joint_current(self, joint_id: Union[int, str]) -> float:
        return 0

    @abstractmethod
    def get_joint_status(self, joint_id: Union[int, str]) -> int:
        return 0

    @abstractmethod
    def lean(self, direction: str, amount: int, move_time: int) -> bool:
        return False

    @abstractmethod
    def walk(self, num_steps: int = 2, start_foot:str = 'auto', turn: int = 0, 
                step_length:int = 40, move_time: int = 1500) -> bool:
        return False

    @abstractmethod
    def eyes(self, joint_id: int, pose_or_angle: Union[str, int], move_time: int = 100) -> bool:
        return False

    @abstractmethod
    def kick(self, side: str = 'right', twist: int = 0, move_time: int = 2000) -> bool:
        return False

    @abstractmethod
    def arms(self, left_angle: int, right_angle: int, move_time: int) -> bool:
        return False

    @abstractmethod
    def celebrate(self, move_time: int = 4000) -> bool:
        return False

    @abstractmethod
    def circle_dance(self, side: str = 'right', move_time: int = 1500) -> bool:
        return False

    @abstractmethod
    def dance(self, side: str = 'right', move_time: int = 1500) -> bool:
        return False

    @abstractmethod
    def wiggle(self, move_time: int = 1500) -> bool:
        return False

    @abstractmethod
    def sidestep(self, side: str, steps: int = 1, step_length: int = 100, 
            move_time: int = 2000) -> bool:
        return False

    @abstractmethod
    def play_sound(self, name_or_freq_start: Union[str,float], 
            freq_end: Optional[float] = None, 
            duration: Optional[int] = None) -> bool:
        return False

    @abstractmethod
    def pinmode_gpio(self, gpio: int, mode: str) -> bool:
        return False

    @abstractmethod
    def write_gpio(self, gpio: int, value: int) -> bool:
        return False

    @abstractmethod
    def digitalread_gpio(self, gpio: int) -> bool:
        return False

    @abstractmethod
    def i2c_write(self, *byte_array: int) -> bool:
        return False

    @abstractmethod
    def i2c_write_to_ric(self, address: int, byte_array: bytes) -> bool:
        return False

    @abstractmethod
    def get_battery_voltage(self) -> float:
        return 0

    @abstractmethod
    def get_battery_remaining(self) -> float:
        return 0

    @abstractmethod
    def get_distance_sensor(self) -> float:
        return 0

    @abstractmethod
    def get_accelerometer(self, axis: Optional[str] = None, axisCode: int = 0) -> float:
        return 0

    @abstractmethod
    def enable_motors(self, enable: bool = True, clear_queue: bool = True) -> bool:
        return False
                  
    @abstractmethod
    def enable_safeties(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def fall_protection(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def motor_protection(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def battery_protection(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def buzz_prevention(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def lifelike_behaviour(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def set_parameter(self, *byte_array: int) -> bool:
        return False

    @abstractmethod
    def save_calibration(self) -> bool:
        return False

    @abstractmethod
    def clear_calibration(self) -> bool:
        return False

    @abstractmethod
    def is_calibrated(self) -> bool:
        return False

    @abstractmethod
    def ros_command(self, *byte_array: int) -> bool:
        return False

    @abstractmethod
    def keyframe (self, time: float, num_of_msgs: int, msgs) -> List[bytes]:
        return False

    @abstractmethod
    def get_chatter(self) -> bytes:
        return False

    @abstractmethod
    def get_firmware_version(self) -> bool:
        return False

    @abstractmethod
    def _mute_serial(self) -> bool:
        return False

    @abstractmethod
    def ros_serial_formatter(self, topicID: int, send: bool = False, *message: int) -> List[int]:
        return False

    @abstractmethod
    def is_moving(self) -> bool:
        return False

    @abstractmethod
    def is_paused(self) -> bool:
        return False

    @abstractmethod
    def get_robot_status(self) -> Dict:
        return {}

    @abstractmethod
    def get_joints(self) -> Dict:
        return {}

    @abstractmethod
    def get_power_status(self) -> Dict:
        return {}

    @abstractmethod
    def get_add_ons_status(self) -> Dict:
        return {}

    @abstractmethod
    def get_add_on_status(self, add_on_name_or_id: Union[int, str]) -> Dict:
        return {}

    @abstractmethod
    def get_system_info(self) -> Dict:
        return {}

    @abstractmethod
    def set_marty_name(self, name: str) -> bool:
        return False

    @abstractmethod
    def get_marty_name(self) -> str:
        return ""

    @abstractmethod
    def is_marty_name_set(self) -> bool:
        return False

    @abstractmethod
    def get_hw_elems_list(self) -> List:
        return []

    @abstractmethod
    def send_ric_rest_cmd(self, ricRestCmd: str) -> None:
        pass

    @abstractmethod
    def send_ric_rest_cmd_sync(self, ricRestCmd: str) -> Dict:
        return {}

    @abstractmethod
    def register_logging_callback(self, loggingCallback: Callable[[str],None]) -> None:
        pass

    @abstractmethod
    def get_interface_stats(self) -> Dict:
        return {}
        
    @abstractmethod
    def preException(self, isFatal: bool) -> None:
        pass

    @abstractmethod
    def get_test_output(self) -> dict:
        return ""



'''
WebSocketFrame
'''
import io
import struct
import secrets
import itertools
import collections
import logging
import time
from typing import Deque

logger = logging.getLogger(__name__)

class WebSocketFrame():

    max_size = 20000
    OPCODE_CONT = 0x00
    OPCODE_TEXT = 0x01
    OPCODE_BINARY = 0x02
    OPCODE_CLOSE = 0x08
    OPCODE_PING = 0x09
    OPCODE_PONG = 0x0a

    def __init__(self) -> None:
        self.curInputData: bytearray = bytearray()
        self.rxFrameData: bytearray = bytearray()
        self.rxFrameIsText = False
        self.pongRequired = False
        self.pongData = bytearray()
        self.closeRequired = False
        self.textMsgs: Deque[str] = collections.deque()
        self.binaryMsgs: Deque[bytes] = collections.deque()
        self.statsPings = 0
        self.statsPongs = 0
        self.statsText = 0
        self.statsBinary = 0

    def getPongRequired(self) -> bool:
        tmpPongReq = self.pongRequired
        self.pongRequired = False
        return tmpPongReq

    def getPongData(self) -> bytes:
        return self.pongData

    def getCloseRequired(self) -> bool:
        return self.closeRequired

    def getTextMsg(self) -> str:
        if len(self.textMsgs) > 0:
            return self.textMsgs.popleft()
        return None

    def getBinaryMsg(self) -> bytes:
        if len(self.binaryMsgs) > 0:
            binaryMsg = self.binaryMsgs.popleft()
            return binaryMsg
        return None

    def addDataToDecode(self, data: bytes) -> None:
        '''
        Add data to be decoded
        Args:
            data bytes to add
        Returns:
            None
        '''
        # Add data
        self.curInputData += data
        # logger.debug(f"{''.join('{:02x}'.format(x) for x in data)}")
        while self._extractFrames():
            pass

    def _extractFrames(self) -> bool:
        if len(self.curInputData) < 2:
            return False
        head1, head2 = struct.unpack("!BB", self.curInputData[:2])

        # Extract header info
        fin = (head1 & 0b10000000) != 0
        opcode = head1 & 0b00001111
        mask = (head2 & 0b10000000) != 0
        frameLen = head2 & 0b01111111
        curPos = 2
        if frameLen == 126:
            if len(self.curInputData) < 4:
                return False
            (frameLen,) = struct.unpack("!H", self.curInputData[2:4])
            curPos = 4
        elif frameLen == 127:
            if len(self.curInputData) < 10:
                return False
            (frameLen,) = struct.unpack("!Q", self.curInputData[2:10])
            curPos = 10
        if frameLen > self.max_size:
            self.curInputData.clear()
            return False
        if mask and len(self.curInputData) < curPos + 4:
            return False
        if mask:
            self.mask_bytes = self.curInputData[curPos:curPos+4]
            curPos += 4

        # Check data is present
        if len(self.curInputData) < curPos + frameLen:
            return False
        rxDataBlock = self.curInputData[curPos : curPos + frameLen]
        if mask:
            rxDataBlock = self.applyMask(rxDataBlock, self.mask_bytes)
        
        # Remove data consumed
        self.curInputData = self.curInputData[curPos + frameLen:]

        # Check for pings
        if opcode == self.OPCODE_PING:
            self.pongRequired = True
            self.pongData = rxDataBlock
            self.statsPings += 1
            # logger.debug(f"wsFrame PING {''.join('{:02x}'.format(x) for x in rxDataBlock)}")
            return True
        elif opcode == self.OPCODE_PONG:
            self.statsPongs += 1
            # logger.debug(f"wsFrame PONG {''.join('{:02x}'.format(x) for x in rxDataBlock)}")
            return True
        elif opcode == self.OPCODE_CLOSE:
            # logger.debug(f"wsFrame CLOSE")
            self.closeRequired = True
            return True
        elif opcode == self.OPCODE_BINARY:
            # logger.debug(f"wsFrame BINARY PART {''.join('{:02x}'.format(x) for x in rxDataBlock)}")
            self.rxFrameIsText = False
        elif opcode == self.OPCODE_TEXT:
            # logger.debug(f"wsFrame TEXT PART {''.join('{:02x}'.format(x) for x in rxDataBlock)}")
            self.rxFrameIsText = True

        # Add data to received frame
        self.rxFrameData += rxDataBlock

        # Add completed frame to queue
        if fin:
            if self.rxFrameIsText:
                self.textMsgs.append(self.rxFrameData.decode('utf-8'))
                self.statsText += 1
            else:
                # logger.debug(f"wsFrame BINARY DONE {''.join('{:02x}'.format(x) for x in self.rxFrameData)}")
                self.binaryMsgs.append(self.rxFrameData[:])
                self.statsBinary += 1
            self.rxFrameData.clear()
        return True

    @classmethod
    def encode(self, inFrame: bytes, useMask: bool,
                opcode: int, fin: bool) -> bytes:

        output = io.BytesIO()

        # Prepare the header.
        head1 = 0b10000000 if fin else 0 | opcode
        head2 = 0b10000000 if useMask else 0
        length = len(inFrame)
        if length < 126:
            output.write(struct.pack("!BB", head1, head2 | length))
        elif length < 65536:
            output.write(struct.pack("!BBH", head1, head2 | 126, length))
        else:
            output.write(struct.pack("!BBQ", head1, head2 | 127, length))

        if useMask:
            mask_bytes = secrets.token_bytes(4)
            output.write(mask_bytes)
            data = self.applyMask(inFrame, mask_bytes)
        else:
            data = inFrame
        output.write(data)
        return output.getvalue()

    @classmethod
    def applyMask(cls, data: bytes, mask: bytes) -> bytes:
        if len(mask) != 4:
            return
        return bytes(b ^ m for b, m in zip(data, itertools.cycle(mask)))



'''
Implements a frame protocol on top of a standard ASCII interface
'''
class ProtocolOverAscii():
    '''
    ProtocolOverAscii
    The main serial port on RIC (USB C connector) is used for plain-text logging AND control
    This protocol makes that possible by assuming bit-7 of any plain-text would be 0 and then
    condensing all control communications into the remaining (bit-7 == 1) values of each byte
    '''
    OVERASCII_ESCAPE_1 = 0x85
    OVERASCII_ESCAPE_2 = 0x8E
    OVERASCII_ESCAPE_3 = 0x8F
    OVERASCII_MOD_CODE = 0x20

    def __init__(self):
        self.escapeSeqCode = 0

    def decodeByte(self, ch: int) -> int:
        '''
        Function to handle char for decoding
        Returns -1 if char was an escape code - remainder will be returned later
        Args:
            ch: byte to decode
        Returns:
            decoded byte or -1 if the byte was an escape code (see above)
        '''
        # Check if in escape sequence
        if self.escapeSeqCode != 0:
            prevEscCode = self.escapeSeqCode
            self.escapeSeqCode = 0
            if prevEscCode == 1:
                return (ch ^ self.OVERASCII_MOD_CODE) & 0x7f
            elif prevEscCode == 2:
                return ch ^ self.OVERASCII_MOD_CODE
            return ch
        elif ch == self.OVERASCII_ESCAPE_1:
            self.escapeSeqCode = 1
            return -1
        elif ch == self.OVERASCII_ESCAPE_2:
            self.escapeSeqCode = 2
            return -1
        elif ch == self.OVERASCII_ESCAPE_3:
            self.escapeSeqCode = 3
            return -1
        else:
            return ch & 0x7f

    @classmethod
    def encode(cls, inData: bytes) -> bytes:
        '''
        Encode frame
        Values 0x00-0x0F map to ESCAPE_CODE_1, VALUE_XOR_20H_AND_MSB_SET
        Values 0x10-0x7f map to VALUE_WITH_MSB_SET
        Values 0x80-0x8F map to ESCAPE_CODE_2, VALUE_XOR_20H
        Values 0x90-0xff map to ESCAPE_CODE_3, VALUE
        Value ESCAPE_CODE_1 maps to ESCAPE_CODE_1 + VALUE
        Args:
            inData: data to encode (bytes)
        Returns:
            encoded frame (bytes)
        '''
        # Iterate over frame
        encodedFrame = bytearray()
        for toEnc in inData:
            # Handle escape chars
            if toEnc <= 0x0f:
                encodedFrame.append(ProtocolOverAscii.OVERASCII_ESCAPE_1)
                encodedFrame.append((toEnc ^ ProtocolOverAscii.OVERASCII_MOD_CODE) | 0x80)
            elif (toEnc >= 0x10) and (toEnc <= 0x7f):
                encodedFrame.append(toEnc | 0x80)
            elif (toEnc >= 0x80) and (toEnc <= 0x8f):
                encodedFrame.append(ProtocolOverAscii.OVERASCII_ESCAPE_2)
                encodedFrame.append(toEnc ^ ProtocolOverAscii.OVERASCII_MOD_CODE)
            else:
                encodedFrame.append(ProtocolOverAscii.OVERASCII_ESCAPE_3)
                encodedFrame.append(toEnc)
        return encodedFrame
